/*
  climate.y - Weather FIGs generating utility / bison part
  Copyright (C) 1999 Michael Jung
  e-mail: miju@phantasia.org

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* ---------------------- */
/* --- C DECLARATIONS --- */
/* ---------------------- */

%{
#include <math.h>
#include <stdio.h>
#include "climate.h"
#define YYDEBUG 0
int yydebug = 1;

/* --- prototypes --- */
  
int error(int id);
int yyerror (char *s);
int yylex(void);

/* --- variables --- */

FILE      *fi, *fo;   /* In-/Out-files */
weather   *wlist;     /* List of all weathers */
table     *tlist;     /* List of weather-table-lists */
calender  *clist;     /* Months list */
int       nwatches;   /* Number of watches per day */
float     light;      /* maximal daylight:day*/
date      *hilight;   /* date of maximal daylight:day */
cycles    *plist;     /* List of periods/cycles */
translate *tllist;    /* List of translations */
char      *prgname;
jumptable *rlist;     /* Jump-to-new-weather list */
%}

/* -------------------------- */
/* --- BISON DECLARATIONS --- */
/* -------------------------- */

%union {
  int       val;
  int       *iptr;
  char      *string;
  timespan  *tsptr;
  date      *dptr;
  weather   *wptr;
  table     *tptr;
  llist     *lptr;
  calender  *cptr;
  cycle     *pptr;
  cycles    *psptr;
  translate *tlptr;
  jump      *jptr;
  jumptable *jtptr;
}

%token CLOUD PRECIP WDIR WSPEED TEMP
%token CALENDAR CLIMATE TABLELIST

%token <val>    NUM
%token <string> STRING

%type <val>    label
%type <string> cloud precip wdir wspeed temp
%type <string> local
%type <tsptr>  timespan
%type <dptr>   date
%type <wptr>   phenomena weather climate
%type <tptr>   tables
%type <lptr>   labels
%type <cptr>   days calendar
%type <pptr>   restcycle
%type <psptr>  cycles
%type <tlptr>  translates
%type <jptr>   nums
%type <iptr>   num
%type <jtptr>  random

%%

/* --------------------- */
/* --- GRAMMAR RULES --- */
/* --------------------- */

file : 
   | CALENDAR calendar CLIMATE climate TABLELIST tables translates random
  { clist = $2; wlist = $4; tlist = $6; tllist = $7; rlist = $8};

random : '(' labels ':' nums ')'
  { $$ = SIZEALLOC(1, jumptable); $$->next = NULL; $$->labels = $2;
  $$->table = $4; };
   | random '(' labels ':' nums ')'
  { $$ = SIZEALLOC(1, jumptable); $$->next = $1; $$->labels = $3;
  $$->table = $5; };

nums : 
  { $$ = NULL; }
   | nums num
  { $$ = SIZEALLOC(1, jump); $$->next = $1; $$->table = $2; };

num : '+' NUM
  { $$ = SIZEALLOC(2,int); $$[0] = $2; $$[1] = 0; }
   | '-' NUM
  { $$ = SIZEALLOC(2,int); $$[0] = -$2; $$[1] = 0; }
   | NUM
  { $$ = SIZEALLOC(2,int); $$[0] = 0; $$[1] = $1; };

calendar : NUM ':' days daylight cycles
  { nwatches = $1; $$ = $3; plist = $5};

cycles :
  { $$ = NULL; }
   | '(' NUM ':' restcycle ')' cycles
  { $$ = SIZEALLOC(1,cycles); $$->iptr = $$->cptr = $4; $$->lvl = $2;
  $$->next = $6; };

restcycle : STRING
  { $$ = SIZEALLOC(1,cycle); $$->next = NULL; $$->name = $1; }
   | STRING restcycle
  { $$ = SIZEALLOC(1,cycle); $$->next = $2; $$->name = $1; };

daylight : date ':' NUM ':' NUM
  { hilight = $1; light = (float) $3 / $5; };

days :
  { $$ = NULL; }
   | days NUM
  { $$ = SIZEALLOC(1, calender); $$->next = $1; $$->ndays = $2; };

tables : 
  { $$ = NULL; }
   | tables local timespan ':' labels
  { $$ = SIZEALLOC(1, table); $$->next = $1; $$->local = $2;
  $$->tsptr = $3; $$->lptr = $5; };

translates :
  { $$ = NULL; }
   | translates '(' labels ')'
  { $$ = SIZEALLOC(1, translate); $$->next = $1; $$->lptr = $3; };

local : STRING
  { $$ = $1; };

timespan : date '-' date
  { $$ = SIZEALLOC(1, timespan); $$->d1 = $1; $$->d2 = $3; };

date : NUM '.' NUM '.' NUM
  { $$ = SIZEALLOC(1, date); $$->w = $1; $$->d = $3; $$->m = $5; };

labels :
  { $$ = NULL; }
   | labels label
  { $$ = SIZEALLOC(1, llist); $$->label = $2; $$->next = $1; };

label : NUM
  { $$ = $1; };

climate :
  { $$ = NULL; }
   | climate weather
  { $$ = $2; $$->next = $1; };

weather : labels ':' phenomena
  { $3->labels = $1; $$ = $3; };

phenomena :
  { $$ = SIZEALLOC(1, weather); $$->cloud = $$->precip = "";
  $$->temp = $$->wdir = $$->wspeed = ""; };
   | phenomena cloud 
  { $$ = $1; $$->cloud = $2; }
   | phenomena precip
  { $$ = $1; $$->precip = $2; }
   | phenomena temp
  { $$ = $1; $$->temp = $2; }
   | phenomena wdir
  { $$ = $1; $$->wdir = $2; }
   | phenomena wspeed
  { $$ = $1; $$->wspeed = $2; };

cloud : CLOUD STRING
  { $$ = $2; };

precip : PRECIP STRING
  { $$ = $2; };

temp : TEMP STRING
  { $$ = $2; };

wdir : WDIR STRING
  { $$ = $2; };

wspeed : WSPEED STRING
  { $$ = $2; };

%%

/* ------------------ */
/* ADDITIONAL C CODE */
/* ----------------- */

int yylex(void)
{
  static char symb[MAXBUFLEN];
  int         i;
  int         c;
  weather     *wptr;

  /* - Ignore formating characters - */

  do { c = getc(fi); } while ( (c == ' ') || (c == '\t') || (c == '\n') );

  /* - End of Input - */

  if (c == EOF) return(0);
 
  /* - Numbers - */

  if ( isdigit(c) ) { ungetc(c, fi); fscanf(fi,"%d", &yylval); return(NUM); }

  /* - Key words - */

  if (isalpha(c))
    {
      i = 0;
      do
	{ symb[i++] = c; c = getc(fi); }
      while ( (c != EOF) && isalnum(c) && (i < MAXBUFLEN) );

      if (i == MAXBUFLEN) error(1);

      ungetc(c, fi);
      symb[i] = '\0';

      if (!strncmp("CLOUD", symb, 5)) return (CLOUD);
      if (!strncmp("PRECIPITATION", symb, 6)) return (PRECIP);
      if (!strncmp("WDIRECTION", symb, 4)) return (WDIR);
      if (!strncmp("WSPEED", symb, 6)) return (WSPEED);
      if (!strncmp("TEMPERATURE", symb, 4)) return (TEMP);
      if (!strncmp("TABLELIST", symb, 9)) return (TABLELIST);
      if (!strncmp("CALENDAR", symb, 9)) return (CALENDAR);
      if (!strncmp("CLIMATE", symb, 9)) return (CLIMATE);
    }

  /* - Strings - */

  if (c == '"')
    {
      i = 0;
      do
	{ c = getc(fi); symb[i++] = c; }
      while ( (c != EOF) && (c != '"') && (i < MAXBUFLEN) );

      if (i == MAXBUFLEN) error(1);
      symb[i-1] = '\0';

      yylval.string = (char *) malloc( strlen(symb) + 1 );
      strcpy(yylval.string, symb);
      return(STRING);
    }

  /* - Comments - */

  if (c == '#')
    {
      do
	{ c = getc(fi); }
      while ( (c != EOF) && (c != '\n') );
      ungetc(c, fi);
      c = yylex();
    }

  return(c);
}

int yyerror(char *s)
{
  fprintf(stderr, "%s\n", s);
}

/* ------------------------------------------------------------------------- */
/*    Preamble                                                               */
/* ------------------------------------------------------------------------- */

/* - table parameters - */

int wi = 0;   /* Weather Index := 1st entry */

/* - internal variables - */

char    *version = "1.0";
int     pwi = 0;
int     out = 0;          /* Form of Output := 0 */
char    fin[255] = "\0";  /* In filename */
char    fon[255] = "\0";  /* Out filename */
char    rin[255] = "\0";  /* Random number filename */
int     test = 0;         /* Test tables */
canvas  fset =            /* FIG settings */
  { 8100, 11880, 0, 0, 0, 0, 0, 0, 6, 0 };
randnum *grn = NULL;

/* - weather variables - */

weather  *wtable;           /* Table of used weathers */
int      tablen;            /* Length of weather table */
calender *ctable;           /* Calender */
int      callen;            /* Calender Length */
char     here[255] = "\0";  /* Local for generation */

date     fd = {1, 1, 1};    /* Timespan for generation */
date     ld = {1, 1, 1};
timespan now = { &fd, &ld, NULL};

int      tdays;             /* number of days of the year */
int      nday;              /* day of the year */
int      odays;             /* offset days when max light occurs */

/* - prototypes - */

int  randi(int n);
void usage(void);
void prepare(void);
int  dayspan(date *d, timespan *ts);
int  spanspan(timespan *ts1, timespan *ts2);
void wmerge(int merge, int i, weather *wptr, timespan *tsptr);
void initout(void);
void wout(date *d, int i);
int  draw(date *d, char *feature);
void textout(date *d, char *decl, char *feature);
void dr_box(int x1, int y1, int x2, int y2, int col,int depth,int frame);
void dr_moon(int cx, int cy, int r, int col);
void dr_sun(int cx, int cy, int p);
void dr_cloud(int cx, int cy, int p);
void dr_rain(int x, int y, int p);
void dr_snow(int cx, int cy, int p);
void dr_fog(int cx, int cy, int w);
void dr_blizzard(int cx, int cy, int p);
void recfree(cycle *cptr);

/*****************************************************************************/

/* ------------------------------------------------------------------------- */
/*    randi                                                                  */
/* ------------------------------------------------------------------------- */

int randi(int n)
{
  int ret;

  if (grn != NULL) {
    if (grn->num != -1) ret = grn->num%n;
    else ret = (rand()%n);
    grn = grn->next;
  }
  else {
    ret = (rand()%n);
  }
  return ret;
}

/* ------------------------------------------------------------------------- */
/*    main                                                                   */
/* ------------------------------------------------------------------------- */

int main(int argc, char *argv[])
{
  date      today;
  int       opt;
  cycles    *tpptr;
  jumptable *jtptr;
  jump      *jptr, *tjptr;
  int       tj;
  llist     *lptr,*lptr2,*ttjptr;
  translate *ttptr;
  int       maxj, j;
  int       start;
  int       tdays;
  FILE      *ri;
  randnum   *rn,*lrn;
  char      c;

  /* - Get options - */
  
  prgname = *argv; argc--; argv++;
  while (argc)
    {
      opt = 0;
      if (!strcmp(*argv, "-o"))        opt = 1;
      if (!strcmp(*argv, "-v"))        opt = 2;
      if (!strcmp(*argv, "-wi"))       opt = 3;
      if (!strcmp(*argv, "-text"))     opt = 4;
      if (!strcmp(*argv, "-local"))    opt = 5;
      if (!strcmp(*argv, "-timespan")) opt = 6;
      if (!strcmp(*argv, "-fig"))      opt = 7;
      if (!strcmp(*argv, "-font"))     opt = 8;
      if (!strcmp(*argv, "-test"))     opt = 9;
      if (!strcmp(*argv, "-i"))        opt = 10;
      if (!strcmp(*argv, "-printwi"))  opt = 11;
      if (!strcmp(*argv, "-seed"))     opt = 12;
      if (!strcmp(*argv, "-rolls"))    opt = 13;

      switch (opt)
	{
	case 1 : argc--; strcpy(fon, *(++argv));
	  break;
	case 2 : fprintf(stdout, "climate version %s\n", prgname, version);
	  exit(0);
	case 3 : argc--; wi = strtol(*(++argv), NULL, 10) - 1;
	  break;
	case 4 : out = 1;
	  break;
	case 5 : argc--; strcpy(here, *(++argv));
	  break;
	case 6 : argc--; sscanf(*(++argv), "%d.%d.%d-%d.%d.%d",
				&(now.d1->w), &(now.d1->d), &(now.d1->m),
				&(now.d2->w), &(now.d2->d), &(now.d2->m));
	  break;
	case 7 : out = 2;
	  break;
	case 8 : argc--; fset.font = strtol(*(++argv), NULL, 10) - 1;
	  break;
	case 9 : test = 1;
	  break;
	case 10: argc--; strcpy(fin, *(++argv));
	  break;
	case 11: pwi = 1;
	  break;
	case 12 : argc--; srand(strtol(*(++argv), NULL, 10));
	  break;
	case 13 : argc--; strcpy(rin, *(++argv));
	  break;
	}

      if (!opt) usage();
      argc--;argv++;
    }
  
  if (fin[0]) fi = fopen(fin, "r"); else fi = stdin;
  if (fon[0]) fo = fopen(fon, "w"); else fo = stdout;
  
  /* - Random numbers - */

  if (rin[0]) {
    ri = fopen(rin, "r");
    rn = NULL;
    x : while ((c = getc(ri)) != EOF) {
      if ( (c == ' ') || (c == '\t') || (c == '\n') ) goto x;
      j = -1;
      if ( isdigit(c) ) { ungetc(c, fi); fscanf(fi,"%d", &j); }
      lrn = (randnum*)malloc(sizeof(randnum));
      lrn->num = j;
      lrn->next = rn;
      rn = lrn;
    }
    grn = rn;
  }
  
  /* - Prepare - */
  
  yyparse();
  prepare();
  initout();

  /* - Generation Loop - */
  
  if (!test)
    {
      today = *now.d1; fset.ndays = 0;

      if ( (wi > tablen) || (!wtable[wi].tsptr) ||
	   (!dayspan(&today, wtable[wi].tsptr)) )
	error(3);
  
      today.w = today.d = today.m = 1;
      start = 0;
      nday = 0;
    }
  else wi = 1;

  if (!test)
    do {
      /* - Initialize cycles - */
      if ((today.w == now.d1->w) && (today.d == now.d1->d) &&
	  (today.m == now.d1->m))
	start = 1;

      /* - Generate and display weather - */
      if ( start && dayspan(&today, &now) )
	{
	  /* get translates */
	  ttptr = tllist;
	  ttjptr = NULL;
	  while ( ttptr && !ttjptr)
	    {
	      lptr = ttptr->lptr;
	      while (lptr && !ttjptr)
		{
		  if ( wi == lptr->label - 1 )
		    ttjptr = ttptr->lptr;
		  lptr = lptr->next;
		}
	      ttptr = ttptr->next;
	    }
          lptr2 = ttjptr;
	  /* translate if necessary */
	  do
	    {
	      jtptr = rlist;
	      jptr = NULL;
	      while ( jtptr && !jptr )
		{
		  lptr = jtptr->labels;
		  while (lptr && !jptr)
		    {
		      if ( dayspan(&today,wtable[wi].tsptr) &&
			   (wi == lptr->label - 1) )
			jptr = jtptr->table;
		      lptr = lptr->next;
		    }
		  jtptr = jtptr->next;
		}
	      if (!(lptr2 && !jptr)) break;
	      {
		wi = lptr2->label - 1;
		lptr2 = lptr2->next;
	      }
	    }
	  while (1);/* lptr2 && !jptr */

	  maxj = 0; tjptr = jptr;
	  while (tjptr) { maxj++; tjptr = tjptr->next; }
	  if (!maxj) error(8);

	  j = randi(maxj); tjptr = jptr;
	  while (j--) { tjptr = tjptr->next; }
	  if (tjptr->table[1]) wi = tjptr->table[1] - 1;
	  else wi += tjptr->table[0];

	  if ((wi < 0) || (wi > tablen)) { error(6); wi = 0; }

	  if (!wtable[wi].tsptr) error(5);
	  else
	    {
	      wout(&today, wi);
	      /* night time */
	      if ( (out == 2) && (today.w == nwatches) )
		draw(&today,"internal-1");
	    }
	}

      /* - time tick - */
      /* Next watch */
      today.w = today.w % nwatches + 1;
      tpptr = plist;
      while (tpptr)
	{
	  if (tpptr->lvl == 1)
	    if (tpptr->cptr->next) tpptr->cptr = tpptr->cptr->next;
	    else tpptr->cptr = tpptr->iptr;
	  tpptr = tpptr->next;
	}
      
      /* Next day */
      if (today.w == 1)
	{
	  nday++;
	  fset.ndays++;
	  today.d = today.d % ctable[today.m - 1].ndays + 1;
	  tpptr = plist;
	  while (tpptr)
	    {
	      if (tpptr->lvl == 2)
		if (tpptr->cptr->next) tpptr->cptr = tpptr->cptr->next;
		else tpptr->cptr = tpptr->iptr;
	      tpptr = tpptr->next;
	    }
	}
      
      /* Next month */
      if ((today.d == 1) && (today.w == 1))
	{
	  today.m = today.m % callen + 1;
	  tpptr = plist;
	  while (tpptr)
	    {
	      if (tpptr->lvl == 3)
		if (tpptr->cptr->next) tpptr->cptr = tpptr->cptr->next;
		else tpptr->cptr = tpptr->iptr;
	      tpptr = tpptr->next;
	    }
	}

    } while ( (today.w != now.d1->w) || (today.d != now.d1->d) ||
	      (today.m != now.d1->m) || !start);
  else
    {
      today.w = today.d = today.m = 1;
      for (wi = 0; wi < tablen; wi++)
	{
	  if (wtable[wi].tsptr) wout(&today,wi);

	  /* Next watch */
	  today.w = today.w % nwatches + 1;
	  tpptr = plist;
	  while (tpptr)
	    {
	      if (tpptr->lvl == 1)
		if (tpptr->cptr->next) tpptr->cptr = tpptr->cptr->next;
		else tpptr->cptr = tpptr->iptr;
	      tpptr = tpptr->next;
	    }
	  
	  /* Next day */
	  if (today.w == 1)
	    {
	      fset.ndays++;
	      today.d = today.d % ctable[today.m - 1].ndays + 1;
	      tpptr = plist;
	      while (tpptr)
		{
		  if (tpptr->lvl == 2)
		    if (tpptr->cptr->next) tpptr->cptr = tpptr->cptr->next;
		    else tpptr->cptr = tpptr->iptr;
		  tpptr = tpptr->next;
		}
	    }
	  
	  /* Next month */
	  if ((today.d == 1) && (today.w == 1))
	    {
	      today.m = today.m % callen + 1;
	      tpptr = plist;
	      while (tpptr)
		{
		  if (tpptr->lvl == 3)
		    if (tpptr->cptr->next) tpptr->cptr = tpptr->cptr->next;
		    else tpptr->cptr = tpptr->iptr;
		  tpptr = tpptr->next;
		}
	    }
	}
    }
  return(0);
}

/* ------------------------------------------------------------------------- */
/*    usage                                                                  */
/* ------------------------------------------------------------------------- */

void usage(void)
{
  fprintf(stderr,"usage: climate -text|-fig -local place [-v] [-font size] [-wi index]\n");
  fprintf(stderr,"  [-timespan period] [-test] [-i file] [-o file] [-printwi]\n");
  fprintf(stderr,"  -text     specifies text output\n");
  fprintf(stderr,"  -fig      specifies FIG output\n");
  fprintf(stderr,"  -local    specifies weather location (default empty)\n");
  fprintf(stderr,"  -v        prints version\n");
  fprintf(stderr,"  -font     specifies FIG font size (default 6)\n");
  fprintf(stderr,"  -wi       specifies initial weather index (default 1)\n");
  fprintf(stderr,"  -timespan gives period for which weather is generated\n");
  fprintf(stderr,"            (default none)\n");
  fprintf(stderr,"  -test     tests the parsed weather-data file, instead of randomly\n");
  fprintf(stderr,"            generating weather passes through all weather\n");
  fprintf(stderr,"  -i        specifies input file (default reads from\n");
  fprintf(stderr,"            stdin)\n");
  fprintf(stderr,"  -o        specifies output file (default writes to\n");
  fprintf(stderr,"            stdout)\n");
  fprintf(stderr,"  -printwi  puts current weather index into output\n");
  fprintf(stderr,"  -seed     random seed\n");
  fprintf(stderr,"  -rolls    specifies random number file\n");
  exit(-1);
};

/* ------------------------------------------------------------------------- */
/*    error                                                                  */
/* ------------------------------------------------------------------------- */

int error(int id)
{
  static int total = 0;
  char *etxt[] =
  {
    "This error cannot appear",              /* 0 */
    "Symbol size exceeded",                  /* 1 */
    "Out of Memory",                         /* 2 */
    "Illegal weather index",                 /* 3 */
    "No valid weather for this period",      /* 4 */
    "Weather index not valid for this day",  /* 5 */
    "Weather Index out of range",            /* 6 */
    "FIG Output out of bounds",              /* 7 */
    "Weather out of bounds"                  /* 8 */
  };

  fprintf(stderr, "%d. Error: %s\n", ++total, etxt[id]);
}

/* ------------------------------------------------------------------------- */
/*    prepare                                                                */
/* ------------------------------------------------------------------------- */

void prepare(void)
{
  int      i,s;
  table    *tptr;
  llist    *lptr;
  weather  *wptr;
  calender *cptr;

  /* - Calendar - */

  callen = 0; cptr = clist;
  while (cptr)
    {
      callen++;
      cptr = cptr->next;
    }
  
  ctable = SIZEALLOC(callen, calender);
  cptr = clist;
  for (i = callen - 1; i >= 0; i--)
    { ctable[i] = *cptr; cptr = cptr->next; }

  /* - Find highest label - */

  tablen = 0; tptr = tlist;
  while(tptr)
    {
      if ( (!strcmp(tptr->local, here)) || test )
      	{
	  lptr = tptr->lptr;
	  while(lptr) {
	    tablen = MAX(tablen, lptr->label);
          lptr = lptr->next; }
	}
      tptr = tptr->next;
    }
  if (tablen == 0) error(4);

  /* - Initialize weather table - */

  wtable = SIZEALLOC(tablen, weather);
  for (i = 0; i < tablen; i++)
    wmerge(0, i, NULL, NULL);

  /* - Get usable weathers - */

  tptr = tlist;
  while(tptr)
    {
      if ( (!strcmp(tptr->local, here)) || test )
	{
	  lptr = tptr->lptr;
	  while(lptr)
	    {
	      wmerge(2, lptr->label - 1, NULL, tptr->tsptr);
	      lptr = lptr->next;
	    }
	}
      tptr = tptr->next;
    }

  /* - Get weather data - */

  wptr = wlist;
  while (wptr)
    {
      lptr = wptr->labels;
      while(lptr)
	{
          if ( (lptr->label - 1 < tablen) && (wtable[lptr->label - 1].tsptr) )
	    wmerge(1, lptr->label - 1, wptr, NULL);
	  lptr = lptr->next;
	}
      wptr = wptr->next;
    }

  /* - Return "days of the year" - */

  tdays = 0;
  for (i = 0; i < callen; i++)
    {
      if (i < hilight->m - 1) odays += ctable[i].ndays;
      if (i == hilight->m - 1) odays += hilight->d - 1;
      tdays += ctable[i].ndays;
    }
}

/* ------------------------------------------------------------------------- */
/*    dayspan                                                                */
/* ------------------------------------------------------------------------- */

int dayspan(date *d, timespan *ts)
{
  float f, fa, fe;

  f = d->m + d->d/100.0 + d->w/1000.0;
  fa = ts->d1->m + ts->d1->d/100.0 + ts->d1->w/1000.0;
  fe = ts->d2->m + ts->d2->d/100.0 + ts->d2->w/1000.0;

  if ( (f - fa) * (fe - f) * (fe - fa) > 0 ) return(1);
  if ( (f - fa) * (fe - f) == 0 ) return(1);

  if (ts->next) return( dayspan(d, ts->next) ); else return(0);
}

/* ------------------------------------------------------------------------- */
/*    spanspan                                                               */
/* ------------------------------------------------------------------------- */

int spanspan(timespan *ts1, timespan *ts2)
{
  return( dayspan(ts1->d1, ts2) || dayspan(ts2->d1, ts1) );
}

/* ------------------------------------------------------------------------- */
/*    wmerge                                                                 */
/* ------------------------------------------------------------------------- */

void wmerge(int merge, int i, weather *wptr, timespan *tsptr)
{
  float    f1, f2, f3;
  timespan *ts;

  if (merge == 0)   /* Initialize */
    {
      wtable[i].cloud = wtable[i].precip = wtable[i].temp = "";
      wtable[i].wdir = wtable[i].wspeed = "";
      wtable[i].next = NULL; wtable[i].labels = NULL;
      wtable[i].tsptr = NULL;
    }
  if (merge == 1)   /* Merge where table overrides */
    {
      if (!strcmp(wtable[i].cloud, "")) wtable[i].cloud = wptr->cloud;
      if (!strcmp(wtable[i].precip, "")) wtable[i].precip = wptr->precip;
      if (!strcmp(wtable[i].wdir, "")) wtable[i].wdir = wptr->wdir;
      if (!strcmp(wtable[i].wspeed, "")) wtable[i].wspeed = wptr->wspeed;
      if (!strcmp(wtable[i].temp, "")) wtable[i].temp = wptr->temp;
    }
  if (merge == 2)   /* Set table entry timespan */
    {
      ts = SIZEALLOC(1, timespan);
      *ts = *tsptr;

      ts->next = wtable[i].tsptr; wtable[i].tsptr = ts;
    }
}

/* ------------------------------------------------------------------------- */
/*    initout                                                                */
/* ------------------------------------------------------------------------- */

void initout(void)
{
  int i, m;

  /* - Text - */

  if (out == 1)
    {
      fprintf(fo, "Weather for %d.%d.%d - %d.%d.%d in %s\n",
              now.d1->w, now.d1->d, now.d1->m, now.d2->w, now.d2->d, now.d2->m,
              here);
      fprintf(fo, "(generated by climate version %s)\n\n", version);
    }

  /* - Fig - */

  if (out == 2)
    {
      /* Initialize */
      m = 1;
      fset.ew = fset.w / nwatches;
      for (i = 0; i < callen; i++) m = MAX(m, ctable[i].ndays);
      fset.eh = fset.h / m;
      fset.ps = MIN(fset.ew / 48, fset.eh / 48);

      /* Standard Entries */
      fprintf(fo, "#FIG 3.1\n");
      fprintf(fo, "#file generated by %s version %s\n", prgname, version);
      fprintf(fo, "Center\nMetric\n");
      fprintf(fo, "1200 2\n");
    }
}

/* ------------------------------------------------------------------------- */
/*    wout                                                                   */
/* ------------------------------------------------------------------------- */

void wout(date *d, int i)
{
  cycles *tpptr;
  cycle  *fptr,*tptr;
  char   *txt;

  /* Cycles */
  tpptr = plist; fptr = NULL;
  while (tpptr)
    {
      if (tpptr->cptr->name[0])
	{
	  tptr = SIZEALLOC(1,cycle);
	  tptr->name = tpptr->cptr->name;
	  tptr->next = fptr;
	  fptr = tptr;
	}
      tpptr = tpptr->next;
    }

  /* - Text - */

  txt = NULL;
  if (out == 1)
    {
      /* Date */
      fprintf(fo, "%d.%d.%d.:", d->w, d->d, d->m);

      /* Cycles */
      tptr = fptr;
      while (tptr)
	{
	  if (tptr == fptr) fprintf(fo, " (%s", tptr->name);
	  else fprintf(fo, ", %s", tptr->name);
	  tptr = tptr->next;
	}
      if (fptr) fprintf(fo, ")\n"); else fprintf(fo, "\n");

      /* wi */
      if (pwi) fprintf(fo, " WEATHER INDEX: %d\n", wi+1);

      /* Phenomena */
      if (wtable[i].cloud[0])
	fprintf(fo, " SKY: %s\n", wtable[i].cloud);
      if (wtable[i].precip[0])
	fprintf(fo, " PRECIP.: %s\n", wtable[i].precip);
      if (wtable[i].wdir[0])
	{
	  fprintf(fo, " WIND: %s ", wtable[i].wdir);
	  if (wtable[i].wspeed[0]) fprintf(fo, "(%s)\n", wtable[i].wspeed);
	  else fprintf(fo, "\n");
	}
      else
	if (wtable[i].wspeed[0]) fprintf(fo, "WIND: (%s)", wtable[i].wspeed);
      if (wtable[i].temp[0]) fprintf(fo, " TEMP: %s\n", wtable[i].temp);
      fprintf(fo, "\n");
    }

  /* - Fig - */

  if (out == 2)
    {
      fset.loff = fset.ruoff = fset.rloff = 0;

      /* Date */
      /* wi */
      if (pwi) fprintf(fo, "#WEATHER INDEX: %d\n", wi);

      /* Cycles */
      tptr = fptr;
      while (tptr)
	{
	  if (!draw(d, tptr->name)) textout(d, "", tptr->name);
	  tptr = tptr->next;
	}

      /* Draw as draw can */
      if (!draw(d, wtable[i].cloud)) textout(d, "CLOUDS:", wtable[i].cloud);
      if (!draw(d, wtable[i].precip)) textout(d, "PRECIP.:", wtable[i].precip);

      if (!draw(d, wtable[i].wdir))
        if (!draw(d, wtable[i].wspeed))
	  if (wtable[i].wdir[0])
	    if (wtable[i].wspeed[0])
	      {
		txt = SIZEALLOC(strlen(wtable[i].wspeed) +
				strlen(wtable[i].wdir) + 4, char);
		sprintf(txt,"%s (%s)", wtable[i].wdir, wtable[i].wspeed);
		textout(d, "WIND:", txt);
	      }
	    else textout(d, "WIND:", wtable[i].wdir);
	  else
	    if (wtable[i].wspeed[0])
	      textout(d, "WINDSPEED:", wtable[i].wspeed);
	    else ;
	else
	  if (wtable[i].wdir[0]) textout(d, "WIND:", wtable[i].wdir);
	  else ;
      else
        if (!draw(d, wtable[i].wspeed))
	  if (wtable[i].wspeed[0]) textout(d, "WINDSPEED:", wtable[i].wspeed);

      if (!draw(d, wtable[i].temp)) textout(d, "TEMP:", wtable[i].temp);
    }

  /* Free */
  recfree(tptr);
  if (txt) free(txt);
}

/* ------------------------------------------------------------------------- */
/*    draw                                                                   */
/* ------------------------------------------------------------------------- */

int draw(date *d, char *feature)
{
  int lox, loy, rux, ruy, p, ok;
  static int firstday = -1;

  lox = fset.ew * d->w + fset.loff;
  if (firstday == -1) firstday = fset.ndays;
  loy = fset.eh * (fset.ndays - firstday + 1);
  rux = lox + fset.ew;
  ruy = loy + fset.eh;
  p = fset.ps;

  ok = 0;
  if (!strcmp(feature,"Day"))   /* Day Box */
    {
      dr_box(lox - fset.loff, loy, rux - fset.loff, ruy, 0, 2, 1);
      ok = 1;
    }
  if (!strcmp(feature,"Night"))   /* Night Box */
    {
      dr_box(lox - fset.loff, loy, rux - fset.loff, ruy, 0, 2, 1);
      ok = 1;
    }
  if (!strcmp(feature,"internal-1"))   /* Night light */
    {
      dr_box(fset.ew, loy, fset.ew * (1+d->w)/2 - fset.ew * d->w/2 *
	     fabs(light-0.5) * sin ( (float) (nday-odays) / tdays * 8*atan(1)
				     + (float) 2*atan(1) ),
	     ruy, 17, 3, 0);
      dr_box(fset.ew * (1+d->w), loy, fset.ew * (0.5+d->w) + fset.ew * d->w/2 *
	     fabs(light-0.5) * sin ( (float) (nday-odays) / tdays * 8*atan(1)
				     + (float) 2*atan(1) ),
	     ruy, 17, 3, 0);
      ok = 1;




    }
  if (!strcmp(feature,"Full moon"))   /* Full moon */
    {
      dr_moon(lox + p*18, loy + p*24, p*10, 3);
      fset.loff += p*30;
      ok = 1;
    }
  if (!strcmp(feature,"New moon"))   /* New moon */
    {
      dr_moon(lox + p*18, loy + p*24, p*10, 0);
      fset.loff += p*30;
      ok = 1;
    }
  if (!strcmp(feature,"clear"))   /* Clear sky */
    {
      dr_sun(lox + p*18, loy + p*18, p);
      fset.loff += p*36;
      ok = 1;
    }
  if (!strcmp(feature,"overcast"))   /* Overcast sky */
    {
      dr_cloud(lox + p*18, loy + p*30, p);
      fset.loff += p*36;
      ok = 1;
    }
  if (!strcmp(feature,"partly"))   /* Overcast sky */
    {
      dr_sun(lox + p*18, loy + p*18, p);
      dr_cloud(lox + p*18, loy + p*30, p);
      fset.loff += p*36;
      ok = 1;
    }
  if (!strcmp(feature,"rain"))   /* Rain */
    {
      dr_rain(lox + p*2, loy + p*10, 2*p);
      fset.loff += p*24;
      ok = 1;
    }
  if (!strcmp(feature,"showers"))   /* Showers */
    {
      dr_rain(lox + p, loy + 7*p, p);
      dr_rain(lox + 13*p, loy + 7*p, p);
      dr_rain(lox + 7*p, loy + 20*p, p);
      fset.loff += p*24;
      ok = 1;
    }
  if (!strcmp(feature,"snow"))   /* Snow */
    {
      dr_snow(lox + 12*p, loy + 20*p, 2*p);
      fset.loff += p*24;
      ok = 1;
    }
  if (!strcmp(feature,"snow flurries"))   /* Flurries */
    {
      dr_snow(lox + 5*p, loy + 15*p, p);
      dr_snow(lox + 20*p, loy + 15*p, p);
      dr_snow(lox + 12*p, loy + 30*p, p);
      fset.loff += p*24;
      ok = 1;
    }
  if (!strcmp(feature,"fog"))   /* Fog */
    {
      dr_fog(lox + 12*p, loy + 24*p, 10*p);
      fset.loff += p*24;
      ok = 1;
    }
  if (!strcmp(feature,"blizzard"))   /* Blizzard */
    {
      dr_blizzard(lox + 12*p, loy + 20*p, 2*p);
      fset.loff += p*24;
      ok = 1;
    }

  if (fset.loff + MAX(fset.ruoff, fset.rloff) >= fset.ew) error(7);
  return (ok);
}

/* ------------------------------------------------------------------------- */
/*    textout                                                                */
/* ------------------------------------------------------------------------- */

void textout(date *d, char *decl, char *feature)
{
  int  rux, ruy, fs, p;
  int  strclen;
  static int firstday = -1;

  if (!feature[0]) return;

  rux = fset.ew * d->w + fset.ew;
  if (firstday == -1) firstday = fset.ndays;
  ruy = fset.eh * (fset.ndays - firstday + 1) + fset.eh;
  fs = fset.font;
  p = fset.ps;

  /*
   * The following line is definitely wrong, but XFIG doesn't seem to
   * mind
   */

  strclen = (strlen(decl) + strlen(feature)) * fs * 6;

  fprintf(fo,"4 2 -1 0 0 -1 %d 0.000 4 %d %d ", fs, 10*fs, strclen);

  if (fset.ruoff < fset.rloff)
    {
      fprintf(fo,"%d %d ", rux - fset.ruoff - 2*p, ruy - 10*fs - 6*p);
      fset.ruoff += strclen;
    }
  else
    {
      fprintf(fo,"%d %d ", rux - fset.rloff - 2*p, ruy - 2*p);
      fset.rloff += strclen;
    }

  if (fset.loff + MAX(fset.ruoff, fset.rloff) >= fset.ew) error(7);
  fprintf(fo,"%s %s\\001\n", decl, feature);
}

/* ------------------------------------------------------------------------- */
/*    dr_box                                                                 */
/* ------------------------------------------------------------------------- */

void dr_box(int x1, int y1, int x2, int y2, int col,int depth,int frame)
{
  fprintf(fo,"2 2 0 %d -1 %d %d 0 30 0.000 1 0 7 0 0 5\n", frame, col, depth);
  fprintf(fo,"%d %d %d %d %d %d %d %d %d %d\n", x1, y1, x1, y2,
	  x2, y2, x2, y1, x1, y1);
}

/* ------------------------------------------------------------------------- */
/*    dr_moon                                                                */
/* ------------------------------------------------------------------------- */

void dr_moon(int cx, int cy, int r, int col)
{
  fprintf(fo,"1 3 0 1 -1 %d 0 0 20 0.000 1 0.000 ", col);
  fprintf(fo,"%d %d %d %d %d %d %d %d\n", cx, cy, r, r, cx, cy,
	  cx - r, cy);
}

/* ------------------------------------------------------------------------- */
/*    dr_sun                                                                 */
/* ------------------------------------------------------------------------- */

void dr_sun(int cx, int cy, int p)
{
  dr_moon(cx, cy - 2*p, 8*p, 7);
}

/* ------------------------------------------------------------------------- */
/*    dr_cloud                                                               */
/* ------------------------------------------------------------------------- */

void dr_cloud(int cx, int cy, int p)
{
  dr_moon(cx - 6*p, cy, 6*p, 0);
  dr_moon(cx + 6*p, cy, 6*p, 0);
  dr_moon(cx, cy - 6*p, 6*p, 0);
  fprintf(fo,"2 2 0 1 -1 0 2 0 20 0.000 1 0 7 0 0 5\n");
  fprintf(fo,"%d %d %d %d %d %d %d %d %d %d\n", cx-6*p, cy-6*p,
	  cx-6*p, cy+6*p, cx+6*p, cy+6*p, cx+6*p, cy-6*p, cx-6*p,
	  cy-6*p);
}

/* ------------------------------------------------------------------------- */
/*    dr_rain                                                                */
/* ------------------------------------------------------------------------- */

void dr_rain(int x, int y, int p)
{
  fprintf(fo,"2 3 0 1 -1 -1 0 0 20 0.000 1 0 -1 0 0 5\n");
  fprintf(fo,"%d %d %d %d %d %d %d %d %d %d\n", x + 10*p, y, x, y + 7*p, x,
	  y + 10*p, x + 3*p, y + 10*p, x + 10*p, y);
}

/* ------------------------------------------------------------------------- */
/*    dr_snow                                                                */
/* ------------------------------------------------------------------------- */

void dr_snow(int cx, int cy, int p)
{
  fprintf(fo,"2 3 0 1 -1 -1 0 0 20 0.000 1 0 -1 0 0 19\n");
  fprintf(fo,"%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d ",
	  cx - 5*p, cy - p, cx - 2*p, cy - p, cx - 4*p, cy - 5*p,
	  cx - 2*p, cy - 5*p, cx, cy - 2*p, cx + 2*p, cy - 5*p,
	  cx + 4*p, cy - 5*p, cx + 2*p, cy - p, cx + 5*p, cy - p,
	  cx + 5*p);
  fprintf(fo,"%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n",
	  cy + p, cx + 2*p, cy + p, cx + 4*p, cy + 5*p, cx + 2*p, cy + 5*p,
	  cx, cy + 2*p, cx - 2*p, cy + 5*p, cx - 4*p, cy + 5*p, cx - 2*p,
	  cy + p, cx - 5*p, cy + p, cx - 5*p, cy - p);
}

/* ------------------------------------------------------------------------- */
/*    dr_fog                                                                 */
/* ------------------------------------------------------------------------- */

void dr_fog(int cx, int cy, int w)
{
  fprintf(fo,"5 1 0 1 -1 -1 0 0 10 0.000 1 0 0 0 ");
  fprintf(fo,"%d %d %d %d %d %d %d %d\n",
	  cx, cy + 5*w/4, cx - 3*w/4, cy - 4*w/4, cx + 3*w/4,
	  cy - 5*w/4, cx + 3*w/4, cy - 4*w/4);
  fprintf(fo,"5 1 0 1 -1 -1 0 0 10 0.000 1 0 0 0 ");
  fprintf(fo,"%d %d %d %d %d %d %d %d\n",
	  cx, cy + 7*w/4, cx - 3*w/4, cy - 2*w/4, cx + 3*w/4,
	  cy - 3*w/4, cx + 3*w/4, cy - 2*w/4);
  fprintf(fo,"5 1 0 1 -1 -1 0 0 10 0.000 1 0 0 0 ");
  fprintf(fo,"%d %d %d %d %d %d %d %d\n",
	  cx, cy + 9*w/4, cx - 3*w/4, cy - 0*w/4, cx + 3*w/4,
	  cy - 1*w/4, cx + 3*w/4, cy - 0*w/4);
}

/* ------------------------------------------------------------------------- */
/*    dr_blizzard                                                            */
/* ------------------------------------------------------------------------- */

void dr_blizzard(int cx, int cy, int p)
{
  fprintf(fo,"2 3 0 1 -1 -1 0 0 20 0.000 1 0 -1 0 0 8\n");
  fprintf(fo,"%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n",
	  cx + 4*p, cy - 3*p, cx, cy - 3*p, cx - 4*p, cy + 2*p,
	  cx - 1*p, cy + 2*p, cx - 6*p, cy + 7*p, cx + 4*p, cy + 2*p,
	  cx,       cy + 2*p, cx + 4*p, cy - 3*p);
}

/* ------------------------------------------------------------------------- */
/*    recfree                                                                */
/* ------------------------------------------------------------------------- */

void recfree(cycle *cptr)
{
  if (cptr)
    if (cptr->next) { recfree(cptr->next); free(cptr); }
    else free(cptr);
}
